//
// Sample showing how to use bump mapping in GLSL programs
//
// Author: Alex V. Boreskoff <alexboreskoff@mtu-net.ru>, <steps3d@narod.ru>
//

#include	"libExt.h"

#ifdef	MACOSX
	#include	<GLUT/glut.h>
#else
	#include	<glut.h>
#endif

#include	<stdio.h>
#include	<stdlib.h>

#include	"libTexture.h"
#include	"Vector3D.h"
#include	"Vector2D.h"
#include	"Vector4D.h"
#include	"boxes.h"

Vector3D	eye   ( 4, 3, 4 );			// camera position
float	 	angle = 0;
Vector3D	rot ( 0, 0, 0 );
int			mouseOldX = 0;
int			mouseOldY = 0;
unsigned	decal1Map;
unsigned	decal2Map;

/////////////////////////////////////////////////////////////////////////////////

void    box ( const Vector3D& pos, const Vector3D& size, unsigned texture )
{
    float   x2 = pos.x + size.x;
    float   y2 = pos.y + size.y;
    float   z2 = pos.z + size.z;

    glBindTexture ( GL_TEXTURE_2D, texture );
    glEnable      ( GL_TEXTURE_2D );

    glBegin ( GL_QUADS );
                                    // front face
        glNormal3f   ( 0, 0, 1 );

        glTexCoord2f ( 0, 0 );
        glVertex3f   ( pos.x, pos.y, z2 );

        glTexCoord2f ( size.x, 0 );
        glVertex3f   ( x2, pos.y, z2 );

        glTexCoord2f ( size.x, size.y );
        glVertex3f   ( x2, y2, z2 );

        glTexCoord2f ( 0, size.y );
        glVertex3f   ( pos.x, y2, z2 );

                                    // back face
        glNormal3f ( 0, 0, -1 );

        glTexCoord2f ( size.x, 0 );
        glVertex3f   ( x2, pos.y, pos.z );

        glTexCoord2f ( 0, 0 );
        glVertex3f   ( pos.x, pos.y, pos.z );

        glTexCoord2f ( 0, size.y );
        glVertex3f   ( pos.x, y2, pos.z );

        glTexCoord2f ( size.x, size.y );
        glVertex3f   ( x2, y2, pos.z );

                                    // left face
        glNormal3f ( -1, 0, 0 );

        glTexCoord2f ( 0, 0 );
        glVertex3f   ( pos.x, pos.y, pos.z );

        glTexCoord2f ( 0, size.z );
        glVertex3f   ( pos.x, pos.y, z2 );

        glTexCoord2f ( size.y, size.z );
        glVertex3f   ( pos.x, y2, z2 );

        glTexCoord2f ( size.y, 0 );
        glVertex3f   ( pos.x, y2, pos.z );

                                    // right face
        glNormal3f ( 1, 0, 0 );

        glTexCoord2f ( 0, size.z );
        glVertex3f   ( x2, pos.y, z2 );

        glTexCoord2f ( 0, 0 );
        glVertex3f   ( x2, pos.y, pos.z );

        glTexCoord2f ( size.y, 0 );
        glVertex3f   ( x2, y2, pos.z );

        glTexCoord2f ( size.y, size.z );
        glVertex3f   ( x2, y2, z2 );

                                    // top face
        glNormal3f ( 0, 1, 0 );

        glTexCoord2f ( 0, size.z );
        glVertex3f   ( pos.x, y2, z2 );

        glTexCoord2f ( size.x, size.z );
        glVertex3f   ( x2, y2, z2 );

        glTexCoord2f ( size.x, 0 );
        glVertex3f   ( x2, y2, pos.z );

        glTexCoord2f ( 0, 0 );
        glVertex3f   ( pos.x, y2, pos.z );

                                    // bottom face
        glNormal3f ( 0, -1, 0 );

        glTexCoord2f ( size.x, size.z );
        glVertex3f   ( x2, pos.y, z2 );

        glTexCoord2f ( 0, size.z );
        glVertex3f   ( pos.x, pos.y, z2 );

        glTexCoord2f ( 0, 0 );
        glVertex3f   ( pos.x, pos.y, pos.z );

        glTexCoord2f ( size.x, 0 );
        glVertex3f   ( x2, pos.y, pos.z );

    glEnd ();
}

void init ()
{
	glClearColor ( 0.0, 0.0, 0.0, 1.0 );
	glEnable     ( GL_DEPTH_TEST );
	glDepthFunc  ( GL_LESS       );

	glHint ( GL_POLYGON_SMOOTH_HINT,         GL_NICEST );
	glHint ( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
}

void	drawFirstInSecond ( void (*first)(), void (*second)() )
{
	glEnable    ( GL_CULL_FACE );
	glEnable    ( GL_DEPTH_TEST );
	glColorMask ( GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE );
	glCullFace  ( GL_BACK );
	
	first ();
	
	glDepthMask   ( GL_FALSE );
	glEnable      ( GL_STENCIL_TEST );
	glStencilOp   ( GL_KEEP, GL_KEEP, GL_INCR );
	glStencilFunc ( GL_ALWAYS, 0, 0 );
	
	second ();
	
	glStencilOp ( GL_KEEP, GL_KEEP, GL_DECR );
	glCullFace  ( GL_FRONT );
	
	second ();
	
	glDepthMask   ( GL_TRUE );
	glColorMask   ( GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE );
	glStencilFunc ( GL_NOTEQUAL, 0, 1 );
	glDisable     ( GL_DEPTH_TEST );
	glCullFace    ( GL_BACK );
	
	first ();
	
	glDisable     ( GL_STENCIL_TEST );
}

void	fixDepth ( void (*draw)() )
{
	glColorMask ( GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE );
	glEnable    ( GL_DEPTH_TEST );
	glDisable   ( GL_STENCIL_TEST );
	glDepthFunc ( GL_ALWAYS );
	
	draw ();
	
	glDepthFunc ( GL_LESS );
}

void	drawA ()
{
	Vector3D	p1 ( -1, -1, -1 );
	
	box ( p1, Vector3D ( 2, 2, 2 ), decal1Map );
}

void	drawB ()
{
	Vector3D	p2 ( 0.2, 0.1, 0 );
	
	box ( p2, Vector3D ( 1.3, 1.3, 1.3 ), decal2Map );
}

void display ()
{
	glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT );

	glPopMatrix        ();

	glMatrixMode   ( GL_MODELVIEW );
	glPushMatrix   ();

    glRotatef    ( rot.x, 1, 0, 0 );
    glRotatef    ( rot.y, 0, 1, 0 );
    glRotatef    ( rot.z, 0, 0, 1 );

	glEnable     ( GL_CULL_FACE );
	glCullFace   ( GL_BACK );
	glDepthFunc  ( GL_LESS );
	
	drawFirstInSecond ( drawA, drawB );
	fixDepth          ( drawB );
	drawFirstInSecond ( drawB, drawA );
		
    glPopMatrix ();

	glutSwapBuffers ();
}

void reshape ( int w, int h )
{
   glViewport     ( 0, 0, (GLsizei)w, (GLsizei)h );
   glMatrixMode   ( GL_PROJECTION );
   											// factor all camera ops into projection matrix
   glLoadIdentity ();
   gluPerspective ( 60.0, (GLfloat)w/(GLfloat)h, 1.0, 60.0 );
   gluLookAt      ( eye.x, eye.y, eye.z,	// eye
					0, 0, 0,				// center
					0.0, 0.0, 1.0 );		// up

   glMatrixMode   ( GL_MODELVIEW );
   glLoadIdentity ();
}

void motion ( int x, int y )
{
    rot.y -= ((mouseOldY - y) * 180.0f) / 200.0f;
    rot.z -= ((mouseOldX - x) * 180.0f) / 200.0f;
    rot.x  = 0;

    if ( rot.z > 360 )
		rot.z -= 360;

	if ( rot.z < -360 )
		rot.z += 360;

    if ( rot.y > 360 )
		rot.y -= 360;

	if ( rot.y < -360 )
        rot.y += 360;

    mouseOldX = x;
    mouseOldY = y;

	glutPostRedisplay ();
}

void mouse ( int button, int state, int x, int y )
{
    if ( state == GLUT_DOWN )
    {
        mouseOldX = x;
        mouseOldY = y;
	}
}

void key ( unsigned char key, int x, int y )
{
	if ( key == 27 || key == 'q' || key == 'Q' )	//	quit requested
    	exit ( 0 );
}

void	animate ()
{
	angle  = 0.004f * glutGet ( GLUT_ELAPSED_TIME );

	glutPostRedisplay ();
}

int main ( int argc, char * argv [] )
{
								// initialize glut
	glutInit            ( &argc, argv );
	glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH | GLUT_STENCIL );
	glutInitWindowSize  ( 500, 500 );


								// create window
	glutCreateWindow ( "CSG subtract demo" );

								// register handlers
	glutDisplayFunc  ( display );
	glutReshapeFunc  ( reshape );
	glutKeyboardFunc ( key     );
	glutMouseFunc    ( mouse   );
	glutMotionFunc   ( motion  );
	glutIdleFunc     ( animate );

	init           ();
	initExtensions ();

	decal1Map = createTexture2D ( true, "../../Textures/oak.bmp" );
	decal2Map = createTexture2D ( false, "../../Textures/16.jpg" );

    glutMainLoop ();

	return 0;
}
